// dmx.c  // ***************************************************************
//
// DMX related files for
// simple DMX512 controller for STM32F042F6 (20 pin TSSOP package)
// based on USB MIDISTREAM and CDC-ABSTRACT CONTROL MODEL
//
// Author and (c)2017-21 Wolfgang Schemmert.
// The copyright claim and GPL-3 licence is restricted to this file
// and the depending files "main.c","main.h","usb-fs-042.c", "userinterface.c"
// and "globaldefines.h".
// Other depending material is published and copyrighted
// by STMicroelectronics and Segger GmbH/Rowley Associates Ltd
//
// Contact: <www.midi-and-more.de>
// A detailled hardware construction
// and operation manual is provided at this website
//
// ****************************************************************************
//      This program is free software: you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation, either version 3 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//      Licence: GPL-3
//
// ****************************************************************************

#include "stm32f0xx.h"
#include "globaldefines.h"

// ************************************

//main.h
extern uint32_t gSYSCOUNT, gLEDCOUNT, gTIMEOUT, gDELAYCOUNT;
extern uint32_t gVid, gPid, gChannel, gLedBlink;
extern void msDelay(uint32_t);
extern void sendUsbByte(uint8_t);
extern void sendUsbString(const uint8_t *);
extern void sendB2(uint8_t);
extern void sendX8(uint8_t);
extern void sendX16(uint16_t);
extern void sendX32(uint32_t);
extern void sendUsbN32(uint32_t);   //sendS32() without trailing SPACE
extern void sendUsbS32(uint32_t);
extern uint16_t pullUsbRxBuf(void);

//usb-fs.c
extern void sendMidiMsg(uint8_t, uint8_t, uint8_t, uint8_t);

//userinterface.c
extern uint32_t parseDeciNum(uint32_t);
extern const uint8_t gcFadeStr[];
extern const uint8_t gcCycleStr[];

// ************************************

uint8_t gdMINIDMX_ON;  //gdNoZeroVelocity,

uint8_t gdDMXBUF[DMX_BUFLEN+2];     //actual DMX transmit buffer
uint8_t gdFINALMEM[DMX_BUFLEN+2];   //final level of fade process
uint8_t gdSTEPMEM[DMX_BUFLEN+2];    //used for control of fade timing
uint8_t gdSUBMEM[DMX_BUFLEN+2];
uint8_t gdRESIMEM[DMX_BUFLEN+2];

uint32_t gdDmxAddr, gdDmxTxIndex;
uint32_t gdFaderBusy, gdMiniDmxOn;
uint32_t gdFadeTime, gdDmxMax;

void startDmxTx(void);
void DmxHeader(void);
void wait16us(void);
void sendDmxByte(void);
void MiniDMX_Interpreter(void);
void MiniDMX_ACK(void);
void setDmxAddress(void);
void setDmxLevel(uint8_t);
void setNextDmxLevel(uint8_t);
void setEqual(uint8_t);
void fade2Preset(void);
void startFader(uint16_t);
void fadeProcess(void);
void msgDmxState(uint32_t);
void readDmxBuf(uint32_t);
void sendAddrLevel(void);
void sendLevel(uint16_t);

// ************************************

// DMX handling
//***************

void startDmxTx(void) // ****************************************************
// TSSOP package uses USART2
// USART2-TxD is PortA2,  DMX-Break (Open Drain Pull Down) is PA0
// **************************************************************************
{
uint32_t k;

    RCC->APB1ENR |=  RCC_APB1ENR_USART2EN ; // USART2 clock enable

    GPIOA->MODER &= ~(GPIO_MODER_MODER2 );	//Set pin 2 to alternate function
    GPIOA->MODER |=   GPIO_MODER_MODER2_1;	// 10=Alternate function mode
    //GPIOA->OTYPER is push-pull by default
    GPIOA->OSPEEDR &= ~(GPIO_OSPEEDR_OSPEEDR2_1);
    GPIOA->OSPEEDR |= GPIO_OSPEEDR_OSPEEDR2_0;  //Medium Speed
    //GPIOA->PUPDR is HiZ by default (except PA13,14,15)

// PA0 is Open Drain and pulls DMX OUT to SPACE during DMX Reset
    //configure PA0 as open drain output, zero drive current
	GPIOA->MODER &= ~GPIO_MODER_MODER0;
	GPIOA->MODER |= GPIO_MODER_MODER0_0 ; //PortA0 general purpose output mode
	GPIOA->OSPEEDR &= ~GPIO_OSPEEDR_OSPEEDR0_1;
	GPIOA->OSPEEDR |= GPIO_OSPEEDR_OSPEEDR0_0;  //Medium Speed
	GPIOA->OTYPER |= GPIO_OTYPER_OT_0;
	GPIOA->BSRR = GPIO_BSRR_BS_0;            //Open Drain High= transistor OFF !

//USART is AF1
    GPIOA->AFR[0] &= 0xFFFFF0FF;
    GPIOA->AFR[0] |= 0x00000100;

    USART2->BRR = 0x00C0; //250,000 Baud: 192 at 48 MHz APB Clock

    USART2->CR2 = USART_CR2_STOP_1;	//=2 Stopbits
    USART2->CR3 = 0;			    //default:no Handshake
    USART2->CR1 = USART_CR1_UE | USART_CR1_TE;
                                    //TX only, 8Bit, no Parity, no Interrupts
	for (k=0;k < (DMX_BUFLEN+1);k++)  {
		gdDMXBUF[k] = 0;
		gdFINALMEM[k] = 0;
		gdSTEPMEM[k] = 0;
		gdSUBMEM[k] = 0;
		gdRESIMEM[k] = 0;
	}
	gdDmxAddr = 0;
//no interrupt is used
}


void DmxHeader(void)  // *****************************************************
// ***************************************************************************
{
uint8_t waitcount;

// Mark before Break min 0 max 1sec Includes transmission time of last DMX byte
	waitcount = 25;					//tested ca. 350 us
	while (waitcount)  {
		wait16us();
		waitcount-- ;
    }

// Space for Break = DMX loop Reset "typically" ca 176 us
// switch Open Drain output
	GPIOA->BSRR = GPIO_BSRR_BR_0;	//Open Drain Low= transistor ON !
 	waitcount = 18;					//tested ca 300us
	while (waitcount)  {
		wait16us();
		waitcount-- ;
	}
// release Open Drain output
    GPIOA->BSRR = GPIO_BSRR_BS_0;	//Open Drain High= transistor OFF !
// Mark after Break min 12 us
    wait16us();						//tested ca 35us
    wait16us();
	
// send DMX Startbyte = 0x00
	while ((USART2->ISR & USART_ISR_TXE) == 0) ;	//modified 13July20
	USART2->TDR = 0;	//set DMX StartBit
// the TXE bit is cleared automatically when TDR is written
	gdDmxTxIndex = 0;               //index 0 of DMX buffer is User DMX addr=1
}


void wait16us(void)   // *******************************************************
// blocking delay ca 16us
// *************************************************************************
{
uint32_t countdown;

	countdown = 100;
    while (countdown)   {
        IWDG->KR = 0x0000AAAA;  //timeout ca. 1-2 seconds
        countdown-- ;
    }
}




void MiniDMX_Interpreter(void)  // ******************************************
// **************************************************************************
{
uint8_t tempBuf[DMX_BUFLEN];
uint16_t recvword, index, packetLen;
uint8_t recvbyte;

    gTIMEOUT = 100;

    while (1)  {
        if (gTIMEOUT == 0) {        //seems to be non-MiniDMX input
            gdMiniDmxOn = 0;        //MiniDmx regarded as inactive
            gLedBlink = 0;
            return;
        }
        recvword = pullUsbRxBuf();
        if (recvword)  {
            recvbyte = (uint8_t)(recvword & 0x00FF);

            if (recvbyte == 0xA2) {
                packetLen = 512;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA1) {
                packetLen = 256;
                break;      // OK, start next stage
            }
            else if (recvbyte == 0xA0) {
                packetLen = 96;
                break;      // OK (but NOT TESTED HERE!), start next stage
            }
            else  {                 //unvalid packet specifier: error
                MiniDMX_ACK();      //send ACK even in case of error
                return;             //but stay in MiniDMX mode
            }
        }
    }

    gdMiniDmxOn = 0xFF;         //MiniDmx regarded as active

//valid MiniDMX packet opened
    if (gLedBlink==0) gLedBlink = 1;    //don't restart, else not blinking
    index = 0;  //starts from 0 LIKE gDmxTxIndex !!

    while (1)  {
        gTIMEOUT = 60;           //60ms timeout per byte

        while (1)  {
            recvword = pullUsbRxBuf();
            if (recvword)  {
                recvbyte = (uint8_t)(recvword & 0x00FF);
                break;          //next byte received successfully
            }

            if(gTIMEOUT == 0)  {
                MiniDMX_ACK();  //send ACK even in case of error
                return;         //but stay in MiniDMX mode
            }

        }
//store bytes temporarily not to disturb DMX transmission in case of error
        tempBuf[index] = recvbyte;
//packet is complete when number of DMX bytes specified by protocol
//plus 1 termination byte (0xA5) received
        if (index >= packetLen) break;
        index++ ;
    }
//byte transfer complete --> next stage:
    if (recvbyte != 0xA5) { //check for correct packet termination
        MiniDMX_ACK();      //send ACK even in case of error
        return;             //but stay in MiniDMX mode
    }
//packet received successfully, store into DMX transmit buffer
    for (index=0; index <= packetLen; index++) {
        gdDMXBUF[index] = tempBuf[index];
        gdFINALMEM[index] = tempBuf[index];       //terminate any fading,too
    }
    MiniDMX_ACK();
}


void MiniDMX_ACK(void)  // **************************************************
// **************************************************************************
{
    sendUsbByte(0x5A);
    sendUsbByte(0xC1);  //send ACK even in case of error
    sendUsbByte(0xA5);  //then better get next actual one
}


void setDmxAddress(void)  // ************************************************
// only get and store gDmxAddr as write pointer to DMX transmit buffer
// **************************************************************************
{
uint32_t number;

    number = parseDeciNum(512);
    if ((number > 512) || (number == 0)) {
        sendUsbByte('?');
        return;
    }
    else gdDmxAddr = number- 1;
}


void setDmxLevel(uint8_t level)  // *****************************************
// **************************************************************************
{
#ifdef DMX_DEBUG
	sendB2('*');
	sendX8(level);
#endif
        if (gdFadeTime == 0)  {
        gdFINALMEM[gdDmxAddr] = level;
        gdDMXBUF[gdDmxAddr] = level;
    }
    else  {
        gdFINALMEM[gdDmxAddr] = level;
        startFader(gdDmxAddr);
    }
}


void setNextDmxLevel(uint8_t level)  // *************************************
// **************************************************************************
{
    if (gdDmxAddr == 511)  {
        sendUsbByte('?');
        return;
    }
    gdDmxAddr++ ;
    setDmxLevel(level);
}


void setEqual(uint8_t number)  // *******************************************
// **************************************************************************
{
uint8_t level;
uint16_t address,k;

    if (number == 0) return;
    level = gdFINALMEM[gdDmxAddr];
       address = gdDmxAddr;
    for (k=0;k<number;k++)  {
        gdDmxAddr++;
        setDmxLevel(level);
        if (gdDmxAddr == 511)  break;
    }
    gdDmxAddr = address;
}


void fade2Preset(void)  // **************************************************
// **************************************************************************
{
int16_t k;

    for (k=0;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        if (gdFadeTime == 0)  gdDMXBUF[k] = gdFINALMEM[k];
        else startFader(k);
    }
}


void startFader(uint16_t address)  // ***************************************
// **************************************************************************
{
uint16_t dividend, quotient;
    
    if (gdFadeTime == 0) return;//precaution,usually inhibited by program flow
    if (gdDMXBUF[address] == gdFINALMEM[address]) return;  //all-ready

    dividend = gdDMXBUF[address] - gdFINALMEM[address];
//calculate absolute difference,
    if (gdDMXBUF[address] > gdFINALMEM[address])
                            dividend = gdDMXBUF[address] - gdFINALMEM[address];
    else dividend =  gdFINALMEM[address] - gdDMXBUF[address];

    quotient = (dividend << 8) / gdFadeTime;
                                    //divisor != 0 guaranteed by program flow
    gdSTEPMEM[address] =(uint8_t)(quotient >> 8);    //HighByte
//gdSTEMPMEM contains the number of full gdDMXBUF steps each /50s  (fast fade)
    gdSUBMEM[address] =(uint8_t)(quotient & 0x00FF); //LowByte
//because the the dividend was multiplied by 256, SUBMEM is a kind of remainder
//this is added to RESIMEM (residual of full step), ie. when RESIMEM overflows,
//a full extra +/-step (depends on fade direction) for DMXBUF gets inserted
//(slow fade), execution of this see fadeProcess().
    gdRESIMEM[address] = 0;
    gdFaderBusy = 1;
}


void fadeProcess(void)  // **************************************************
//this is the "ballistic" algorithm (target is computed at start time), with
//very small CPU overhead at runtime. It is less precise for long time fades
//with small differences between actual and final level.
//an alternative would be the "cruise missile" algorithm, which calculates the
//difference between actual and target every /50s and is more precise,
//but takes more CPU time (a division per DMX channel update).
// **************************************************************************
{
uint16_t k, residual;
int16_t actual, temp;
uint8_t step;

    gdFaderBusy = 0;                //temporary,check if further fade is active

    for (k=1;k<DMX_BUFLEN;k++)  {
        if (gdDMXBUF[k] == gdFINALMEM[k]) continue;
        gdFaderBusy = 0xFF;                 //trigger next fadeProcess()
        actual = (uint16_t)gdDMXBUF[k];
        step = gdSTEPMEM[k];
        residual = (uint16_t)gdRESIMEM[k];
        residual += gdSUBMEM[k];
        gdRESIMEM[k] = (uint8_t)(residual & 0x00FF);
        if (residual >= 0x0100) step++; //handle overflow without side effect
        if (actual < (uint16_t)gdFINALMEM[k])  {     //fade UP
            temp = (int16_t)(gdDMXBUF[k]) + step;
            if (temp > gdFINALMEM[k]) actual = gdFINALMEM[k];
            else actual += step;
        }
        else  {                                 //fade DOWN
            temp = (int16_t)(gdDMXBUF[k]) - step;
            if (temp < gdFINALMEM[k]) actual = gdFINALMEM[k];
            else  actual -= step;
        }
        gdDMXBUF[k] = (uint8_t)actual;
    }
}

void msgDmxState(uint32_t mode)  // *****************************************
// **************************************************************************
{
uint8_t timeH;
const uint8_t addressStr [] = "Act.DMX channel:";
const uint8_t channelStr [] = "        channel:";
const uint8_t finalStr [] = "Final:";

	if (mode) sendUsbByte(CRR);
	sendUsbString(addressStr);
	sendUsbS32(gdDmxAddr+1);
	sendUsbByte('T');
	sendUsbByte('x');
	sendUsbByte(':');
	sendUsbS32(gdDMXBUF[gdDmxAddr]);
	sendUsbString(finalStr);
	sendUsbS32(gdFINALMEM[gdDmxAddr]);
	sendUsbByte(' ');
	sendUsbByte(0x28);		//'('
	sendUsbString(gcFadeStr);
	timeH = gdFadeTime/50;
	sendUsbN32(timeH);
	sendUsbByte('.');
	sendUsbN32((gdFadeTime/5) - (timeH*10));
	sendUsbByte('s');
	sendUsbByte(' ');
	sendUsbString(gcCycleStr);
	sendUsbN32(gdDmxMax);
	sendUsbByte(0x29);		//')'
	if (mode)  {
		sendUsbByte(CRR);
		if (gdDmxAddr < 510)  {
			sendUsbString(channelStr);
			sendUsbS32(gdDmxAddr+2);
			sendUsbByte('T');
			sendUsbByte('x');
			sendUsbByte(':');
			sendUsbS32(gdDMXBUF[gdDmxAddr+1]);
			sendUsbString(finalStr);
			sendUsbS32(gdFINALMEM[gdDmxAddr+1]);
			sendUsbByte(CRR);
			sendUsbString(channelStr);
			sendUsbS32(gdDmxAddr+3);
			sendUsbByte('T');
			sendUsbByte('x');
			sendUsbByte(':');
			sendUsbS32(gdDMXBUF[gdDmxAddr+2]);
			sendUsbString(finalStr);
			sendUsbS32(gdFINALMEM[gdDmxAddr+2]);
			sendUsbByte(CRR);
		}
	}
}


void readDmxBuf(uint32_t count)  // *****************************************
// **************************************************************************
{
uint16_t address;
uint32_t number;

	if (count == 0)  {
		sendUsbByte('e');
		sendUsbByte('a');
		sendUsbByte('d');
		number = parseDeciNum(DMX_BUFLEN);
		if ((number > DMX_BUFLEN) || (number== 0))  {
			sendUsbByte('?');
			return;
		}

		sendUsbByte(' ');
		sendUsbByte('f');
		sendUsbByte('r');
		sendUsbByte('o');
		sendUsbByte('m');
		sendUsbByte(' ');
	}
	else number = count;
    sendUsbByte('S');
    sendUsbN32(gdDmxAddr+1);
    sendUsbByte(':');
    sendUsbByte('V');
    sendUsbByte('=');
    address = gdDmxAddr;
    sendUsbN32((int16_t)gdDMXBUF[address]);
    address++ ;
    number -- ;
    while (1)  {
        if ((address == DMX_BUFLEN) || (number== 0)) break;
        sendUsbByte(',');
        sendUsbN32((int16_t)gdDMXBUF[address]);
        address++;
        number--;
    }
}


void sendAddrLevel(void)  // *************************************************
//first return actually addressed DMX channel as Control Change 0x50...0x53
//then return this actual DMX level as Control Change 0x54 or 0x55
// **************************************************************************
{
uint8_t data1, data2;

    if (gdDmxAddr == 511)  {			//gets responded as DMX channel no.0
        data1 = 0x50;
        data2 = 0;
    }
    else  {
       data1 = (uint8_t)((gdDmxAddr + 1)>>7) + 0x50;
							//internally gdDmxAddr is counted from 0 to 511
       data2 =(uint8_t)((gdDmxAddr + 1) & 0x007F);
    }

//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x00FF));	//Status
    sendUsbByte(data1);
    sendUsbByte(data2);
//append the actual Fade Time
//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x00FF));	//Status
    sendUsbByte(1);										//Data1
    sendUsbByte(gdFadeTime / 5);						//Data2

    sendLevel(gdDmxAddr);
}

void sendLevel(uint16_t address)  // ****************************************
//return DMX level of the addressed DMX channel as Control Change 0x54 or 0x55
// **************************************************************************
{
//"sendUsbByte()" calls "enCodeMidi()":
    sendUsbByte(0xB0 | (uint8_t)(gChannel & 0x000000FF));	//Status
    sendUsbByte((uint8_t)(gdDMXBUF[address] & 1) + 0x54);	//Data1
    sendUsbByte((uint8_t)(gdDMXBUF[address] >>1));			//Data2
}

// ##########################################################################



